import hmac
import json
import urllib.parse

import requests.auth
from requests import PreparedRequest as P
from requests import Response as R

from . import base


class AuthBase(requests.auth.AuthBase):

    def __call__(self, r: P):
        raise NotImplementedError('Auth hooks must be callable.')

    @staticmethod
    def check_credentials(**credentials):
        for name, key in credentials.items():
            if not key:
                raise ValueError('{} Key and Secret are needed!')

    @staticmethod
    def url_query_split(url: str):
        _url = url.rsplit('?', 1)[0]
        query = urllib.parse.urlsplit(url).query
        return _url, query

    @staticmethod
    def parse_data(value):
        return urllib.parse.parse_qs(value)

    @staticmethod
    def load_json(value):
        return json.loads(value) if value else {}

    @staticmethod
    def encode_data(payload):
        return urllib.parse.urlencode(payload, doseq=True)

    @staticmethod
    def encode_json(payload):
        return json.dumps(payload).encode()


class ApiKeyAuth(AuthBase):
    """Attaches API KEY Authentication to the given Request object."""

    api_key_param: str = 'api_key'

    def __init__(self,
                 api_key: str,
                 api_key_param: str=None):
        # Set credentials
        self.api_key: str = api_key
        # Override defaults
        if api_key_param is not None:
            self.api_key_param = api_key_param

    def add_api_key(self, r: P):
        # Add the API key as a query parameter
        url, query = self.url_query_split(r.url)
        params = self.parse_data(query)
        params[self.api_key_param] = self.api_key
        r.prepare_url(url, params)

    def auth_request(self, r: P):
        """Authenticate the Request"""
        self.add_api_key(r)

    def __call__(self, r: P):
        self.auth_request(r)
        return r

    def __eq__(self, other):
        return all([
            self.api_key == getattr(other, 'api_key', None),
        ])

    def __ne__(self, other):
        return not self == other


class HMACAuth(AuthBase):
    """Attaches HMAC Authentication to the given Request object."""

    api_key_header: str = 'x-auth-key'
    nonce_header: str = 'x-auth-nonce'
    signature_header: str = 'x-auth-signature'
    content_type_header: str = 'content-type'
    signature_delimiter: str = '\n'
    algorithm = 'sha256'
    timestamp: base.Timestamp = base.timestamp

    def __init__(self,
                 api_key: str,
                 secret: str,
                 api_key_header: str=None,
                 nonce_header: str=None,
                 signature_header: str=None,
                 algorithm=None):
        # Set credentials
        self.api_key: str = api_key
        self.secret: str = secret
        # Override defaults
        if api_key_header is not None:
            self.api_key_header = self.api_key_header
        if nonce_header is not None:
            self.nonce_header = self.nonce_header
        if signature_header is not None:
            self.signature_header = self.signature_header
        if algorithm is not None:
            self.algorithm = self.algorithm
        # Set counters
        self.last_nonce: int = 0
        self.num_401_calls: int = 0

    def _nonce(self):
        return self.timestamp.microseconds()

    def new_nonce(self, nonce: int=None):
        nonce = nonce or self._nonce()
        self.last_nonce = nonce
        return str(nonce)

    def add_api_key(self, r: P):
        r.headers[self.api_key_header] = self.api_key

    def add_nonce(self, r: P, nonce: str):
        r.headers[self.nonce_header] = nonce

    def add_signature(self, r: P, nonce: str):
        message = self.build_message(r, nonce)
        signature = self.sign(message)
        r.headers[self.signature_header] = signature

    def build_message(self, r: P, nonce: str):
        """Build the message to sign"""
        components = [r.method, nonce, r.path_url]
        if r.body:
            components.append(r.body)
        message = self.signature_delimiter.join(components)
        return message

    def sign(self, msg: str):
        """Sign the message"""
        encoded_msg = msg.encode() if isinstance(msg, str) else msg

        h = hmac.new(key=self.secret.encode(),
                     msg=encoded_msg,
                     digestmod=self.algorithm)

        signature = h.hexdigest()

        return signature

    def auth_request(self, r: P, nonce: str):
        """Authenticate the Request"""
        self.add_api_key(r)
        self.add_nonce(r, nonce)
        self.add_signature(r, nonce)

    def handle_redirect(self, r: R, **kwargs):
        """Reset num_401_calls counter on redirects."""
        if r.is_redirect:
            self.num_401_calls = 1

    def handle_401(self, r: R, **kwargs):
        """Takes the given response and re-tries auth with a new nonce."""

        if r.status_code == 401 and self.num_401_calls < 2:

            self.num_401_calls += 1

            # Renew nonce
            nonce = self.new_nonce(max(self._nonce(), self.last_nonce + 1))

            # Consume content and release the original connection
            # to allow our new request to reuse the same one.
            r.content
            r.close()
            prep = r.request.copy()
            cookies = prep._cookies
            requests.auth.extract_cookies_to_jar(cookies, r.request, r.raw)
            prep.prepare_cookies(cookies)

            self.auth_request(prep, nonce)

            _r = r.connection.send(prep, **kwargs)
            _r.history.append(r)
            _r.request = prep

            return _r

        return r

    def __call__(self, r: P):
        # Get nonce and authenticate the Request
        nonce = self.new_nonce()
        self.auth_request(r, nonce)

        # Register hooks
        r.register_hook('response', self.handle_401)
        r.register_hook('response', self.handle_redirect)

        return r

    def __eq__(self, other):
        return all([
            self.api_key == getattr(other, 'api_key', None),
            self.secret == getattr(other, 'secret', None)
        ])

    def __ne__(self, other):
        return not self == other



from collections import namedtuple


class Price(
    namedtuple('price', [
        'quantity',
        'vwap',
        'price',
        'fees',
        'total',
        'json',
    ])
):
    @classmethod
    def create_from_json(cls, price):
        return cls(
            quantity=float(price['quantity']),
            vwap=float(price['vwap']),
            price=float(price['price']),
            fees=float(price['fees']),
            total=float(price['total']),
            json=price,
        )


class OrderBookEntry(
    namedtuple('book_entry', [
        'price',
        'amount',
        'exchange',
    ])
):
    @classmethod
    def create_from_json(cls, book_entry):
        return cls(
            price=float(book_entry[0]),
            amount=float(book_entry[1]),
            exchange=book_entry[2],
        )


class OrderBook(
    namedtuple('order_book', [
        'bids',
        'asks',
        'json',
    ])
):
    @classmethod
    def create_from_json(cls, order_book):
        return cls(
            bids=[OrderBookEntry.create_from_json(book_entry)
                  for book_entry in order_book['bids']],
            asks=[OrderBookEntry.create_from_json(book_entry)
                  for book_entry in order_book['asks']],
            json=order_book,
        )



from collections import namedtuple
from datetime import datetime


def parse_datetime(datetime_str):
    if datetime_str:
        return datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S.%fZ')


class Price(
    namedtuple('price', [
        'quantity',
        'vwap',
        'price',
        'fees',
        'total',
        'json',
    ])
):
    @classmethod
    def create_from_json(cls, price):
        return cls(
            quantity=float(price['quantity']),
            vwap=float(price['vwap']),
            price=float(price['price']),
            fees=float(price['fees']),
            total=float(price['total']),
            json=price,
        )


class OrderBookEntry(
    namedtuple('book_entry', [
        'price',
        'amount',
        'orders',
    ])
):
    @classmethod
    def create_from_json(cls, book_entry):
        return cls(
            price=book_entry['price'],
            amount=book_entry['amount'],
            orders=book_entry['orders'],
        )


class OrderBook(
    namedtuple('order_book', [
        'bids',
        'asks',
        'timestamp',
        'last_price',
        'json',
    ])
):
    @classmethod
    def create_from_json(cls, order_book):
        return cls(
            bids=[OrderBookEntry.create_from_json(book_entry)
                  for book_entry in order_book['bids']],
            asks=[OrderBookEntry.create_from_json(book_entry)
                  for book_entry in order_book['asks']],
            timestamp=order_book['timestamp'],
            last_price=order_book['last_price'],
            json=order_book,
        )


class Trade(
    namedtuple('trade', [
        'id',
        'uuid',
        'engine_id',
        'pair',
        'amount',
        'price',
        'created_at',
    ])
):
    @classmethod
    def create_from_json(cls, trade):
        return cls(
            id=trade['id'],
            uuid=trade['uuid'],
            engine_id=trade['engine_id'],
            pair=trade['pair'],
            amount=trade['amount'],
            price=float(trade['price']),
            created_at=parse_datetime(trade['created_at']),
        )


class Trades(
    namedtuple('trades', [
        'count',
        'next',
        'previous',
        'results',
        'json',
    ])
):
    @classmethod
    def create_from_json(cls, trades):
        return cls(
            count=trades['count'],
            next=trades['next'],
            previous=trades['previous'],
            results=[Trade.create_from_json(trade)
                     for trade in trades['results']],
            json=trades,
        )


class Rates(
    namedtuple('rates', [
        'base',
        'rates',
        'json',
    ])
):
    @classmethod
    def create_from_json(cls, rates):
        return cls(
            base=rates['base'],
            rates=rates['rates'],
            json=rates,
        )



from . import constants as _c
from . import models as _m
from ..base import Client, ModelMixin


class SFOXPublic(Client, ModelMixin):
    """API Doc: https://www.sfox.com/developers"""
    base_url = 'https://api.sfox.com/v1/'

    def best_price(self, side: str, amount: str):
        """Return the price needed for a limit order to execute fully."""
        data = self.get(f'offer/{side}', params={
            'amount': amount
        })
        if self.return_json:
            return data
        return _m.Price.create_from_json(data)

    def best_buy_price(self, amount: str):
        """Return the price needed for a limit BUY order to execute fully."""
        return self.best_price(_c.Side.BUY.value, amount)

    def best_sell_price(self, amount: str):
        """Return the price needed for a limit SELL order to execute fully."""
        return self.best_price(_c.Side.SELL.value, amount)

    def order_book_raw(self):
        return self.get('markets/orderbook')

    def order_book(self, market_making: bool=False):
        """Return the blended order book of all the available exchanges."""
        data = self.order_book_raw()
        if market_making:
            data = data['market_making']
        order_book = {'bids': data['bids'], 'asks': data['asks']}
        if self.return_json:
            return order_book
        return _m.OrderBook.create_from_json(order_book)

    def market_making_order_book(self):
        """Return the blended market making order book of all the available
        exchanges."""
        return self.order_book(market_making=True)

    def exchanges(self):
        """Return all the available exchanges."""
        data = self.order_book_raw()
        return data['exchanges']



from . import constants as _c
from . import models as _m
from .client_auth import CryptoMKTAuth
from .client_public import CryptoMKTPublic

__all__ = [
    'CryptoMKT',
]


class CryptoMKT:
    # Models
    models = _m
    # Enum Types
    Currency = _c.Currency
    Market = _c.Market
    OrderType = _c.OrderType
    # Clients
    Auth = CryptoMKTAuth
    Public = CryptoMKTPublic



from ..base import Client


class BitstampPublic(Client):
    base_url = 'https://www.bitstamp.net/api/'
    error_keys = ['error', 'reason']

    @staticmethod
    def _endpoint_for(endpoint, version=2):
        # NOTE: Bitstamp urls MUST end with /
        if version == 1:
            endpoint = f'{endpoint}/'
        else:
            endpoint = f'v{version}/{endpoint}/'
        return endpoint.lower()

    def ticker(self, currency_pair: str):
        """
        Returns ticker dictionary.
        """
        endpoint = self._endpoint_for(f'ticker/{currency_pair}')
        return self.get(endpoint)

    def ticker_hour(self, currency_pair: str):
        """
        Returns dictionary of the average ticker of the past hour.
        """
        endpoint = self._endpoint_for(f'ticker_hour/{currency_pair}')
        return self.get(endpoint)

    def order_book(self, currency_pair: str):
        """
        Returns dictionary with "bids" and "asks".

        Each is a list of open orders and each order is represented as a list
        of price and amount.
        """
        endpoint = self._endpoint_for(f'order_book/{currency_pair}')
        return self.get(endpoint)

    def transactions(self, currency_pair: str, time_interval: str=None):
        """
        Returns transactions for the last 'timedelta' seconds.

        Parameter time is specified by one of two values of TransRange class.
        """
        params = {'time': str(time_interval)} if time_interval else None
        endpoint = self._endpoint_for(f'transactions/{currency_pair}')
        return self.get(endpoint, params=params)

    def trading_pairs_info(self):
        """
        Returns list of dictionaries specifying details of each available
        trading pair:

            {
                'description':'Litecoin / U.S. dollar',
                'name':'LTC/USD',
                'url_symbol':'ltcusd',
                'trading':'Enabled',
                'minimum_order':'5.0 USD',
                'counter_decimals':2,
                'base_decimals':8
            },
        """
        return self.get(self._endpoint_for('trading-pairs-info'))

    def conversion_rate_usd_eur(self):
        """
        Returns simple dictionary:

            {'buy': 'buy conversion rate', 'sell': 'sell conversion rate'}
        """
        endpoint = self._endpoint_for('eur_usd', version=1)
        return self.get(endpoint)



import base64

from requests import PreparedRequest as P

from . import constants as _c
from . import models as _m
from .client_public import BudaPublic
from ..auth import HMACAuth
from ..base import AuthMixin


class BudaHMACAuth(HMACAuth):

    api_key_header = 'X-SBTC-APIKEY'
    nonce_header = 'X-SBTC-NONCE'
    signature_header = 'X-SBTC-SIGNATURE'
    signature_delimiter = ' '
    algorithm = 'sha384'

    def build_message(self, r: P, nonce: str):
        components = [r.method, r.path_url]
        if r.body:
            encoded_body = base64.b64encode(r.body).decode()
            components.append(encoded_body)
        components.append(nonce)
        message = self.signature_delimiter.join(components)
        return message


class BudaAuth(BudaPublic, AuthMixin):
    auth_cls = BudaHMACAuth

    def __init__(self,
                 key: str,
                 secret: str,
                 timeout: int=None,
                 **kwargs):
        super().__init__(timeout, **kwargs)
        self.add_auth(key, secret)

    def quotation(self,
                  market_id: str,
                  quotation_type: str,
                  amount: float,
                  limit: float=None):
        data = self.post(f'markets/{market_id}/quotations', json={
            'quotation': {
                'type': str(quotation_type),
                'amount': str(amount),
                'limit': str(limit) if limit else None,
            },
        })
        if self.return_json:
            return data
        return _m.Quotation.create_from_json(data['quotation'])

    def quotation_market(self,
                         market_id: str,
                         quotation_type: str,
                         amount: float):
        return self.quotation(market_id, quotation_type, amount, limit=None)

    def quotation_limit(self,
                        market_id: str,
                        quotation_type: str,
                        amount: float,
                        limit: float):
        return self.quotation(market_id, quotation_type, amount, limit)

    # BALANCES-----------------------------------------------------------------
    def balance(self, currency: str):
        data = self.get(f'balances/{currency}')
        if self.return_json:
            return data
        return _m.Balance.create_from_json(data['balance'])

    def balance_event_pages(self,
                            currencies: list,
                            event_names: list,
                            page: int=None,
                            per_page: int=None,
                            relevant: bool=None):
        data = self.get('balance_events', params={
            'currencies[]': [str(c) for c in currencies],
            'event_names[]': [str(e) for e in event_names],
            'page': page,
            'per': per_page,
            'relevant': relevant,
        })
        if self.return_json:
            return data
        # TODO: Response only contains a 'total_count' field instead of meta
        return _m.BalanceEventPages.create_from_json(
            data['balance_events'], data['total_count'], page)

    # ORDERS ------------------------------------------------------------------
    def new_order_payload(self, market_id: str, payload):
        data = self.post(f'markets/{market_id}/orders', json=payload)
        if self.return_json:
            return data
        return _m.Order.create_from_json(data['order'])

    def new_order(self,
                  market_id: str,
                  order_type: str,
                  price_type: str,
                  amount: float,
                  limit: float=None):
        return self.new_order_payload(market_id, payload={
            'type': str(order_type),
            'price_type': str(price_type),
            'amount': str(amount),
            'limit': str(limit) if limit else None,
        })

    def order_pages(self,
                    market_id: str,
                    page: int=None,
                    per_page: int=None,
                    state: str=None,
                    minimum_exchanged: float=None):
        if per_page and per_page > _c.ORDERS_LIMIT:
            msg = f"Param 'per_page' must be <= {_c.ORDERS_LIMIT}"
            raise ValueError(msg)
        data = self.get(f'markets/{market_id}/orders', params={
            'per': per_page,
            'page': page,
            'state': str(state) if state else None,
            'minimum_exchanged': minimum_exchanged,
        })
        if self.return_json:
            return data
        return _m.OrderPages.create_from_json(data['orders'], data.get('meta'))

    def order_details(self, order_id: int):
        data = self.get(f'orders/{order_id}')
        if self.return_json:
            return data
        return _m.Order.create_from_json(data['order'])

    def cancel_order(self, order_id: int):
        data = self.put(f'orders/{order_id}', json={
            'state': _c.OrderState.CANCELING.value,
        })
        if self.return_json:
            return data
        return _m.Order.create_from_json(data['order'])

    # PAYMENTS ----------------------------------------------------------------
    def _transfers(self, path, model, key,
                   page: int=None,
                   per_page: int=None,
                   state: str=None):
        if per_page and per_page > _c.TRANSFERS_LIMIT:
            msg = f"Param 'per_page' must be <= {_c.TRANSFERS_LIMIT}"
            raise ValueError(msg)
        data = self.get(path, params={
            'per': per_page,
            'page': page,
            'state': str(state) if state else None,
        })
        if self.return_json:
            return data
        return model.create_from_json(data[key], data.get('meta'))

    def withdrawal_pages(self,
                         currency: str,
                         page: int=None,
                         per_page: int=None):
        return self._transfers(path=f'currencies/{currency}/withdrawals',
                               model=_m.WithdrawalPages, key='withdrawals',
                               page=page, per_page=per_page)

    def withdrawals(self,
                    currency: str,
                    page: int=None,
                    per_page: int=None):
        data = self.withdrawal_pages(currency, page, per_page)
        if isinstance(data, dict):
            return data['withdrawals']
        return data.withdrawals

    def deposit_pages(self,
                      currency: str,
                      page: int=None,
                      per_page: int=None):
        return self._transfers(path=f'currencies/{currency}/deposits',
                               model=_m.DepositPages, key='deposits',
                               page=page, per_page=per_page)

    def deposits(self,
                 currency: str,
                 page: int=None,
                 per_page: int=None):
        data = self.deposit_pages(currency, page, per_page)
        if isinstance(data, dict):
            return data['deposits']
        return data.deposits

    def withdrawal(self,
                   currency: str,
                   amount: float,
                   target_address: str=None,
                   amount_includes_fee: bool=True,
                   simulate: bool=False):
        data = self.post(f'currencies/{currency}/withdrawals', json={
            'withdrawal_data': {
                'target_address': target_address,
            },
            'amount': str(amount),
            'currency': str(currency),
            'simulate': simulate,
            'amount_includes_fee': amount_includes_fee,
        })
        if self.return_json:
            return data
        return _m.Withdrawal.create_from_json(data['withdrawal'])

    def simulate_withdrawal(self,
                            currency: str,
                            amount: float,
                            amount_includes_fee: bool=True):
        return self.withdrawal(
            currency, amount, target_address=None,
            amount_includes_fee=amount_includes_fee, simulate=True)



from . import __version__
from .bitcoinity import *
from .bitex import *
from .bitfinex import *
from .bitstamp import *
from .buda import *
from .coindesk import *
from .coinmarketcap import *
from .cryptomkt import *
from .currencylayer import *
from .errors import *
from .kraken import *
from .oxr import *
from .ripio import *
from .sfox import *
from .surbtc import *



from datetime import datetime

from . import constants as _c
from . import models as _m
from ..base import Client, ModelMixin


class BudaPublic(Client, ModelMixin):
    base_url = 'https://www.buda.com/api/v2/'
    error_keys = ['message']

    def markets(self):
        data = self.get('markets')
        if self.return_json:
            return data
        return [_m.Market.create_from_json(market)
                for market in data['markets']]

    def market_details(self, market_id: str):
        data = self.get(f'markets/{market_id}')
        if self.return_json:
            return data
        return _m.Market.create_from_json(data['market'])

    def ticker(self, market_id: str):
        data = self.get(f'markets/{market_id}/ticker')
        if self.return_json:
            return data
        return _m.Ticker.create_from_json(data['ticker'])

    def order_book(self, market_id: str):
        data = self.get(f'markets/{market_id}/order_book')
        if self.return_json:
            return data
        return _m.OrderBook.create_from_json(data['order_book'])

    def trades(self,
               market_id: str,
               timestamp: int=None,
               limit: int=None):
        data = self.get(f'markets/{market_id}/trades', params={
            'timestamp': timestamp,
            'limit': limit,
        })
        if self.return_json:
            return data
        return _m.Trades.create_from_json(data['trades'])

    def _report(self,
                market_id: str,
                report_type: _c.ReportType,
                start_at: datetime=None,
                end_at: datetime=None):
        if isinstance(start_at, datetime):
            start_at = int(start_at.timestamp())
        if isinstance(end_at, datetime):
            end_at = int(end_at.timestamp())
        data = self.get(f'markets/{market_id}/reports', params={
            'report_type': str(report_type),
            'from': start_at,
            'to': end_at,
        })
        return data

    def report_average_prices(self,
                              market_id: str,
                              start_at: datetime=None,
                              end_at: datetime=None):
        data = self._report(market_id, _c.ReportType.AVERAGE_PRICES,
                            start_at, end_at)
        if self.return_json:
            return data
        return [_m.AveragePrice.create_from_json(report)
                for report in data['reports']]

    def report_candlestick(self,
                           market_id: str,
                           start_at: datetime=None,
                           end_at: datetime=None):
        data = self._report(market_id, _c.ReportType.CANDLESTICK,
                            start_at, end_at)
        if self.return_json:
            return data
        return [_m.Candlestick.create_from_json(report)
                for report in data['reports']]



from collections import namedtuple


class TradingTicker(
    namedtuple('trading_ticker', [
        'BID',
        'BID_SIZE',
        'ASK',
        'ASK_SIZE',
        'DAILY_CHANGE',
        'DAILY_CHANGE_PERC',
        'LAST_PRICE',
        'VOLUME',
        'HIGH',
        'LOW',
    ])
):

    @classmethod
    def create_from_json(cls, ticker):
        return cls(
            BID=ticker[0],
            BID_SIZE=ticker[1],
            ASK=ticker[2],
            ASK_SIZE=ticker[3],
            DAILY_CHANGE=ticker[4],
            DAILY_CHANGE_PERC=ticker[5],
            LAST_PRICE=ticker[6],
            VOLUME=ticker[7],
            HIGH=ticker[8],
            LOW=ticker[9],
        )


class FoundingTicker(
    namedtuple('founding_ticker', [
        'FRR',
        'BID',
        'BID_SIZE',
        'BID_PERIOD',
        'ASK',
        'ASK_SIZE',
        'ASK_PERIOD',
        'DAILY_CHANGE',
        'DAILY_CHANGE_PERC',
        'LAST_PRICE',
        'VOLUME',
        'HIGH',
        'LOW',
    ])
):

    @classmethod
    def create_from_json(cls, ticker):
        return cls(
            FRR=ticker[0],
            BID=ticker[1],
            BID_SIZE=ticker[2],
            BID_PERIOD=ticker[3],
            ASK=ticker[4],
            ASK_SIZE=ticker[5],
            ASK_PERIOD=ticker[6],
            DAILY_CHANGE=ticker[7],
            DAILY_CHANGE_PERC=ticker[8],
            LAST_PRICE=ticker[9],
            VOLUME=ticker[10],
            HIGH=ticker[11],
            LOW=ticker[12],
        )


class TradingTrade(
    namedtuple('trading_trade', [
        'ID',
        'MTS',
        'AMOUNT',
        'PRICE',
    ])
):

    @classmethod
    def create_from_json(cls, trade):
        return cls(
            ID=trade[0],
            MTS=trade[1],
            AMOUNT=trade[2],
            PRICE=trade[3],
        )


class FoundingTrade(
    namedtuple('founding_trade', [
        'ID',
        'MTS',
        'AMOUNT',
        'RATE',
        'PERIOD',
    ])
):

    @classmethod
    def create_from_json(cls, trade):
        return cls(
            ID=trade[0],
            MTS=trade[1],
            AMOUNT=trade[2],
            RATE=trade[3],
            PERIOD=trade[4],
        )


class TradingBook(
    namedtuple('trading_book', [
        'PRICE',
        'COUNT',
        'AMOUNT',
    ])
):

    @classmethod
    def create_from_json(cls, book):
        return cls(
            PRICE=book[0],
            COUNT=book[1],
            AMOUNT=book[2],
        )


class FoundingBook(
    namedtuple('founding_book', [
        'RATE',
        'PERIOD',
        'COUNT',
        'AMOUNT',
    ])
):

    @classmethod
    def create_from_json(cls, book):
        return cls(
            RATE=book[0],
            PERIOD=book[1],
            COUNT=book[2],
            AMOUNT=book[3],
        )


class Stat(
    namedtuple('stat', [
        'MTS',
        'VALUE',
    ])
):

    @classmethod
    def create_from_json(cls, stat):
        return cls(
            MTS=stat[0],
            VALUE=stat[1],
        )


class Candle(
    namedtuple('candle', [
        'MTS',
        'OPEN',
        'CLOSE',
        'HIGH',
        'LOW',
        'VOLUME',
    ])
):

    @classmethod
    def create_from_json(cls, candle):
        return cls(
            MTS=candle[0],
            OPEN=candle[1],
            CLOSE=candle[2],
            HIGH=candle[3],
            LOW=candle[4],
            VOLUME=candle[5],
        )



import json
from urllib.parse import urlsplit

from requests import PreparedRequest as P

from . import constants as _c
from . import models as _m
from .client_public import CryptoMKTPublic
from ..auth import HMACAuth
from ..base import AuthMixin


class CryptoMKTHMACAuth(HMACAuth):

    api_key_header = 'X-MKT-APIKEY'
    nonce_header = 'X-MKT-TIMESTAMP'
    signature_header = 'X-MKT-SIGNATURE'
    signature_delimiter = ''
    algorithm = 'sha384'

    def _nonce(self):
        return self.timestamp.seconds()

    def build_message(self, r: P, nonce: str):
        path = urlsplit(r.path_url).path
        body = json.loads(r.body) if r.body else {}
        components = [nonce, path]
        for key in sorted(body.keys()):
            value = str(body[key])
            components.append(value)
        message = self.signature_delimiter.join(components)
        return message


class CryptoMKTAuth(CryptoMKTPublic, AuthMixin):
    auth_cls = CryptoMKTHMACAuth

    def __init__(self,
                 key: str,
                 secret: str,
                 timeout: int=None,
                 **kwargs):
        super().__init__(timeout, **kwargs)
        self.add_auth(key, secret)

    # BALANCE------------------------------------------------------------------
    def balance(self):
        data = self.get('balance')
        if self.return_json:
            return data
        return _m.Balance.create_from_json(data['data'])

    def wallet_balance(self, currency: str):
        balance = self.balance()
        return getattr(balance, str(currency))

    # ORDERS-------------------------------------------------------------------
    def active_orders(self,
                      market_id: str,
                      page: int=None,
                      limit: int=_c.ORDERS_LIMIT):
        data = self.get('orders/active', params={
            'market': str(market_id),
            'page': page,
            'limit': limit,
        })
        if self.return_json:
            return data
        return _m.Orders.create_from_json(data['data'], data['pagination'])

    def executed_orders(self,
                        market_id: str,
                        page: int=None,
                        limit: int=_c.ORDERS_LIMIT):
        data = self.get('orders/executed', params={
            'market': str(market_id),
            'page': page,
            'limit': limit,
        })
        if self.return_json:
            return data
        return _m.Orders.create_from_json(data['data'], data['pagination'])

    def create_order(self,
                     market_id: str,
                     order_type: str,
                     amount: float,
                     price: float):
        data = self.post('orders/create', json={
            'market': str(market_id),
            'type': str(order_type),
            'amount': amount,
            'price': price,
        })
        if self.return_json:
            return data
        return _m.Order.create_from_json(data['data'])

    def order_status(self, order_id: str):
        data = self.get('orders/status', params={
            'id': order_id,
        })
        if self.return_json:
            return data
        return _m.Order.create_from_json(data['data'])

    def cancel_order(self, order_id: str):
        data = self.get('orders/cancel', params={
            'id': order_id,
        })
        if self.return_json:
            return data
        return _m.Order.create_from_json(data['data'])

    # PAYMENTS-----------------------------------------------------------------
    # TODO: Not tested
    def create_payment(self,
                       amount: float,
                       currency: str,
                       account_email: str,
                       external_id: str=None,
                       callback_url: str=None,
                       error_url: str=None,
                       success_url: str=None,
                       refund_email: str=None):
        data = self.post('payment/new_order', json={
            'to_receive': amount,
            'to_receive_currency': str(currency),
            'payment_receiver': account_email,
            'external_id': external_id,
            'callback_url': callback_url,
            'error_url': error_url,
            'success_url': success_url,
            'refund_email': refund_email,
        })
        if self.return_json:
            return data
        return data

    # TODO: Not tested
    def payment_status(self, payment_id: str):
        data = self.get('payment/status', params={
            'id': payment_id,
        })
        return data



from requests import Response

from requests.exceptions import RequestException


class APIException(RequestException):
    pass


class InvalidResponse(APIException):
    def __init__(self, error_msg: str, r: Response):
        # Add status code
        message = str(r.status_code)
        # Check for client or server errors
        if 400 <= r.status_code < 500:
            message = f'{message} Client Error:'
        elif 500 <= r.status_code < 600:
            message = f'{message} Server Error:'
        # Add reason
        message = f'{message} {r.reason}'
        # Add message from source
        if error_msg:
            message = f'{message} ({error_msg})'
        # Add url
        message = f'{message} for url: {r.url}'

        super().__init__(message, response=r)
        self.message = message


class DecodeError(APIException):
    def __init__(self, msg, r: Response):
        super().__init__(msg, response=r)



from datetime import date, datetime, timedelta


def clean_empty(d: (dict, list)):
    if not isinstance(d, (dict, list)):
        return d
    if isinstance(d, list):
        return [v for v in (clean_empty(v) for v in d) if v]
    return {k: v for k, v in ((k, clean_empty(v)) for k, v in d.items()) if v}


def date_range(start_date, end_date):
    for n in range(int((end_date - start_date).days)):
        yield start_date + timedelta(n)


def current_utc_date():
    return datetime.utcnow().date()


def format_date_iso(date_value):
    if date is None:
        return None
    if isinstance(date_value, datetime):
        date_value = date_value.date()
    if isinstance(date_value, date):
        date_value = date_value.isoformat()
    return date_value


def format_datetime_iso(date_value):
    if date is None:
        return None
    if isinstance(date_value, date):
        date_value = datetime(
            date_value.year, date_value.month, date_value.day)
    if isinstance(date_value, datetime):
        date_value = date_value.isoformat()
    return date_value



from collections import namedtuple


class Ticker(
    namedtuple('ticker', [
        'last',
        'price_before_last',
        'open',
        'high',
        'low',
        'vwap',
        'volume',
        'bid',
        'ask',
        'json',
    ])
):

    @classmethod
    def create_from_json(cls, ticker):
        return cls(
            last=ticker['last'],
            price_before_last=ticker['price_before_last'],
            open=ticker['open'],
            high=ticker['high'],
            low=ticker['low'],
            vwap=ticker['vwap'],
            volume=ticker['volume'],
            bid=ticker['bid'],
            ask=ticker['ask'],
            json=ticker,
        )


class OrderBookEntry(
    namedtuple('book_entry', [
        'price',
        'amount',
    ])
):

    @classmethod
    def create_from_json(cls, book_entry):
        return cls(
            price=float(book_entry[0]),
            amount=float(book_entry[1]),
        )


class OrderBook(
    namedtuple('order_book', [
        'asks',
        'bids',
        'json',
    ])
):

    @classmethod
    def create_from_json(cls, order_book):
        return cls(
            asks=[OrderBookEntry.create_from_json(entry)
                  for entry in order_book['asks']],
            bids=[OrderBookEntry.create_from_json(entry)
                  for entry in order_book['bids']],
            json=order_book,
        )


class Transaction(
    namedtuple('transaction', [
        'timestamp',
        'id',
        'price',
        'amount',
    ])
):

    @classmethod
    def create_from_json(cls, transaction):
        return cls(
            timestamp=transaction[0],
            id=transaction[1],
            price=transaction[2],
            amount=transaction[3],
        )



from . import models as _m
from ..base import Client, ModelMixin


class RipioExchangePublic(Client, ModelMixin):
    """API Doc: https://www.ripio.com/docs/#ripio-exchange"""
    base_url = 'https://exchange.ripio.com/api/v1/'
    error_keys = ['detail']

    def order_books(self):
        """Fetch order books for all markets"""
        data = self.get('book/')
        if self.return_json:
            return data
        return {market: _m.OrderBook.create_from_json(book)
                for market, book in data.items()}

    def order_book(self, market: str):
        """Fetch order book for the provided market"""
        data = self.order_books()
        return data[market]


# TODO: Ripio Auth not implemented
class RipioExchangeAuth(RipioExchangePublic):

    def trades(self, page: int=None):
        """Fetch last trades"""
        data = self.get('trades/', params={
            'page': page,
        })
        if self.return_json:
            return data
        return _m.Trades.create_from_json(data)


class RipioPublic(Client, ModelMixin):
    """API Doc: https://www.ripio.com/docs/#transactions"""
    base_url = 'https://ripio.com/api/v1/'
    error_keys = ['detail']

    def __init__(self, timeout: int=None, **kwargs):
        super().__init__(timeout, **kwargs)
        self.exchange = RipioExchangePublic(timeout, **kwargs)

    def rates_raw(self):
        return self.get('rates/')

    def rates(self):
        """Fetch rates"""
        data = self.rates_raw()
        data = {'base': data['base'], 'rates': data['rates']}
        if self.return_json:
            return data
        return _m.Rates.create_from_json(data)

    def variation(self):
        """Fetch rates variation"""
        data = self.rates_raw()
        return data['variation']



from . import models as _m
from ..base import Client, ModelMixin


class BitexPublic(Client, ModelMixin):
    """Bitex API Doc: https://bitex.la/developers"""
    base_url = 'https://bitex.la/api-v1/rest/'
    error_keys = ['error']

    def ticker(self, market_id: str):
        """Overview of current market prices and trade volume."""
        data = self.get(f'{market_id}/market/ticker')
        if self.return_json:
            return data
        return _m.Ticker.create_from_json(data)

    def order_book(self, market_id: str):
        """Return bids and asks represented as a list of price and amount."""
        data = self.get(f'{market_id}/market/order_book')
        if self.return_json:
            return data
        return _m.OrderBook.create_from_json(data)

    def _transactions(self, market_id: str, endpoint: str):
        data = self.get(f'{market_id}/market/{endpoint}')
        if self.return_json:
            return data
        return [_m.Transaction.create_from_json(tx) for tx in data]

    def transactions(self, market_id: str):
        """
        Return a list representing all individual trades for the past 60
        minutes, sorted by descending date.
        """
        return self._transactions(market_id, 'transactions')

    def transactions_archive(self, market_id: str):
        """
        Return a list representing all individual trades that took place
        since Bitex started trading until the beginning of the current hour,
        sorted by descending date.

        This is a large download, so Bitex only allows to retrieve it a few
        times per hour. Don't worry though, it only changes once an hour.
        """
        return self._transactions(market_id, 'transactions_archive')



from ..base import Client


class Bitcoinity(Client):
    base_url = 'https://bitcoinity.org/'
    timeout = 15

    def ticker(self,
               currency: str,
               exchange: str,
               span: str):
        return self.get('markets/get_ticker', params={
            'currency': currency,
            'exchange': exchange,
            'span': span,
        })



from . import constants as _c
from . import models as _m
from .client_public import SFOXPublic

__all__ = [
    'SFOX',
]


class SFOX:
    # Models
    models = _m
    # Enum Types
    Side = _c.Side
    # Clients
    Public = SFOXPublic



from . import constants as _c
from .client_auth import BitstampAuth
from .client_public import BitstampPublic

__all__ = [
    'Bitstamp',
]


class Bitstamp:
    # Enum Types
    CurrencyPair = _c.CurrencyPair
    TimeInterval = _c.TimeInterval
    # Clients
    Auth = BitstampAuth
    Public = BitstampPublic



from datetime import datetime

from . import constants as _c
from . import models as _m
from ..base import Client, ModelMixin


class CryptoMKTPublic(Client, ModelMixin):
    base_url = 'https://api.cryptomkt.com/v1/'
    error_keys = ['message']

    def markets(self):
        data = self.get('market')
        return data['data']

    def ticker(self, market_id: str):
        data = self.get('ticker', params={
            'market': str(market_id)
        })
        if self.return_json:
            return data
        return _m.Ticker.create_from_json(data['data'])

    def order_book(self,
                   market_id: str,
                   order_type: str,
                   page: int=None,
                   limit: int=_c.ORDERS_LIMIT):
        data = self.get('book', params={
            'market': str(market_id),
            'type': str(order_type),
            'page': page,
            'limit': limit
        })
        if self.return_json:
            return data
        return _m.OrderBook.create_from_json(
            data['data'], data['pagination'])

    def trades(self,
               market_id: str,
               start: datetime=None,
               end: datetime=None,
               page: int=None,
               limit: int=_c.ORDERS_LIMIT):
        if isinstance(start, datetime):
            start = start.strftime('%Y-%m-%d')
        if isinstance(end, datetime):
            end = end.strftime('%Y-%m-%d')
        data = self.get('trades', params={
            'market': str(market_id),
            'start': start,
            'end': end,
            'page': page,
            'limit': limit,
        })
        if self.return_json:
            return data
        return _m.Trades.create_from_json(
            data['data'], data.get('pagination'))



from . import constants as _c
from .client_auth import KrakenAuth
from .client_public import KrakenPublic

__all__ = [
    'Kraken',
]


class Kraken:
    # Enum Types
    Currency = _c.Currency
    Symbol = _c.Symbol
    # Clients
    Auth = KrakenAuth
    Public = KrakenPublic
